(******************************************************************************
 *                             Patdiff                                        *
 *                                                                            *
 * Copyright (C) 2011- Jane Street Holding, LLC                               *
 *    Contact: opensource@janestreet.com                                      *
 *    WWW: http://www.janestreet.com/ocaml                                    *
 *                                                                            *
 *                                                                            *
 * This library is free software; you can redistribute it and/or              *
 * modify it under the terms of the GNU Lesser General Public                 *
 * License as published by the Free Software Foundation; either               *
 * version 2 of the License, or (at your option) any later version.           *
 *                                                                            *
 * This library is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
 * Lesser General Public License for more details.                            *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public           *
 * License along with this library; if not, write to the Free Software        *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
 *                                                                            *
 ******************************************************************************)

open Core.Std



module Configuration = struct

  let ansi_config = "\
;; -*- scheme -*-
;; patdiff Configuration file

(
 (context 6)

 (line_same
  ((prefix ((text \" |\") (style ((bg bright_black) (fg black)))))))

 (line_old
  ((prefix ((text \"-|\") (style ((bg red)(fg black)))))
   (style ((fg red)))
   (word_same (dim))))

 (line_new
  ((prefix ((text \"+|\") (style ((bg green)(fg black)))))
   (style ((fg green)))))

 (line_unified
  ((prefix ((text \"!|\") (style ((bg yellow)(fg black)))))))

 (header_old
  ((prefix ((text \"------ \") (style ((fg red)))))
   (style (bold))))

 (header_new
  ((prefix ((text \"++++++ \") (style ((fg green)))))
   (style (bold))))


 (hunk
  ((prefix ((text \"@|\") (style ((bg bright_black) (fg black)))))
   (suffix ((text \" ============================================================\") (style ())))
   (style (bold))))
)"

end

module Readme : sig val readme : string end = struct

  module Formatted_text = struct

    module A = Ansi_terminal.ANSITerminal

    let b s = A.apply_string [A.Bold] s
    let u s = A.apply_string [A.Underlined] s

    let name = b "NAME"
    let synopsis = b "SYNOPSIS"
    let patdiff = Filename.basename Sys.argv.(0)
    let patdiff_b = b patdiff
    let description = b "DESCRIPTION"
    let patdiff_u = u patdiff
    let from_file_u = u "from-file"
    let to_file_u = u "to-file"
    let options = b "OPTIONS"
    let subcommands = b "SUBCOMMANDS"
    let compare_b = b "compare"
    let readme_b = b "readme"
    let make_config_b = b "make-config"
    let help_b = b "help"
    let configuration = b "CONFIGURATION"
    let cmd_line_args = b "Command Line Arguments"
    let format_and_styling = b "Format and Styling"

  end

  module F = Formatted_text

  let header = sprintf "\
%s
        patdiff - find differences between two files

%s
        %s [options] from-file to-file

%s
        Put simply, %s compares the contents of the two files
        %s and %s.  If desired, a configuration file defining how
        to format the output can be used."
    F.name F.synopsis F.patdiff_b F.description
    F.patdiff_u F.from_file_u F.to_file_u

  let compare_subcmd = sprintf "\
        %s
        Compare two files using the patience diff algorithm

        By default, compare two files and output the result of the comparison
        to the terminal, highlighting differences between the files on a
        word-by-word basis.  All whitespace is considered equal and
        effectively ignored.  The styling of the output is determined by the
        configuration file used.  A configuration file is searched for at
        ~/.patdiff, and if none is found, default styling is applied.

        The exit code follows the diff standard, much like GNU diff.
        (0 = same, 1 = different, 2 = error)

        options:

        -file FILE
                Use FILE as configuration file (overrides ~/.patdiff)

        -default
                Use the default configuration file (overrides ~./patdiff)

        -context NUM
                Display NUM lines of unchanged context before and after
                changes

        -unrefined
                Don't highlight differences between words, only highlight
                differences between lines (improves performance)

        -keep-whitespace
                Consider whitespace when comparing files

        -latex
                Output in LaTeX format instead of ANSI terminal

        -dont_produce_unified_lines
                When refining and highlighting the differences between the
                words of the files, don't unify the word difference into
                a single unified line.

        -quiet
                Only report whether files differ, don't print the actual
                differences.

        -shallow
                When comparing directories, don't recurse into subdirectories.
                Simply compare the files and directories in the immediate
                directories.  If two subdirectories have the same name but
                different contents, enabling this option will cause them to be
                reported as such but treated as the same.

        -double-check
                If files seem the same, double check with 'cmp'.  If they are
                not actually identical, report so, but still exit 0.

        -mask-uniques
                When comparing directories, don't diff unique files against
                /dev/null.  This is useful if you know the contents of the
                unique files and don't need to see the contents of the entire
                file.

        -ext-cmp FILE
                By default, this program uses OCaml's String.compare function
                to compare each line.  With this option, you can provide an
                external program that takes two anonymous strings as arguments
                and returns exit codes according the diff standard (0 = same,
                1 = different, 2 = error).  %s will then use the
                external program when comparing lines.  Using this option
                implies -unrefined, because the external calls are expensive
                and refinement basically loses its meaning with an arbitrary
                compare	function.

        -alt-old NAME
                In the header of the output, use NAME instead of the default
                old filename and time last modified.

        -alt-new NAME
                Same as -alt-old, but for the new filename."
    F.compare_b F.patdiff_u

  let make_config_subcmd = sprintf "\
        %s FILE
        Write the default configuration file to FILE."
    F.make_config_b

  let readme_subcmd = sprintf "\
        %s
        Output this README text"
    F.readme_b

  let help_subcmd = sprintf "\
        %s
        Output a brief summary of a subcommand, detailing its usage"
    F.help_b

  let subcommands = sprintf "\
%s
        Below is a summary of all of the subcommands that %s accepts.
        Options can be abbreviated to any unique prefix of their name.
        Multiple options cannot be combined into a single command line
        word; they must be separated by spaces.

%s

%s

%s

%s"
    F.subcommands F.patdiff_u compare_subcmd
    make_config_subcmd readme_subcmd help_subcmd

  let configuration = sprintf "\
%s

        Formatting for the output of patdiff is specified in the configuration
        file.  All fields are optional.  The following fields are available:

        config_path
        Directs patdiff to load another configuration file instead of this
        one.  All other options in this file will be ignored.

                (config_path \"/home/username/.patdiff2\")

        %s
        All command line arguments can be passed through the configuration
        file.  If an argument is specified in both the configuration file and
        the command line, the command line overrides the configuration file.

                (context 3)
                (unrefined true)
                (word_unify true)
                (keep_whitespace true)
                (quiet true)
                (shallow true)
                (double_check true)
                (mask_uniques true)
                (latex true)
                (alt_old \"old\")
                (alt_new \"new\")
                (ext_cmp \"str_cmp.sh\")

        %s
        The following styles are available:

            Bold Underline Emph
                (Emph is Underline in ANSI, italics in LaTeX)

            Dim Blink Inverse Hide
                These options mean nothing in LaTeX, and some terminals don't
                support them well, if at all.

            (Fg color) (Foreground color) (Bg color) (Background color)
                Background colors mean nothing in LaTeX.  The following colors
                are available for both LaTeX and ANSI outputs:

                Yellow Blue Black Red Green Magenta Cyan White Gray Default
                Bright_yellow  Bright_blue Bright_black Bright_red
                Bright_magenta Bright_cyan Bright_white Bright_green

                The folowing colors have meaning for LaTeX only:

                CMYK (f,f,f,f)
                    Each f is a number in the range [0-1.00], specifying the
                    percentage of cyan, magenta, yellow, and black respectively
                    in the color.

        Most formatting options consist of three fields: prefix, suffix, and
        style.

        The prefix and suffix fields have identical specifications:

                text
                The characters that will be printed when the prefix or suffix
                is used.

                style
                The styling applied to the prefix or suffix text.

        style
        The styling applied to the contents of the line.


        Line format options have one additional field: word_same

        word_same
        When refining this kind of line, the styling applied to words that are
        the same between the two lines.

        The following formatting options are available:

                line_same
                line_old
                line_new
                line_unified
                word_same_old
                word_same_new
                word_same_unified
                word_old
                word_new
                hunk
                header_old
                header_new

        See the default configuration file to for a sample configuration."
    F.configuration F.cmd_line_args F.format_and_styling

  let readme = sprintf "\
%s

%s

%s"
    header subcommands configuration

end
